@@ -523,12 +523,9 @@ atspi_accessible_get_relation_set (AtspiAccessible *obj, GError **error)
   dbus_message_iter_recurse (&iter, &iter_array);
   while (dbus_message_iter_get_arg_type (&iter_array) != DBUS_TYPE_INVALID)
   {
-    GArray *new_array;
     AtspiRelation *relation;
     relation = _atspi_relation_new_from_iter (&iter_array);
-    new_array = g_array_append_val (ret, relation);
-    if (new_array)
-      ret = new_array;
+    ret = g_array_append_val (ret, relation);
     dbus_message_iter_next (&iter_array);
   }
   dbus_message_unref (reply);
@@ -763,7 +760,7 @@ atspi_accessible_get_toolkit_name (AtspiAccessible *obj, GError **error)
 
   if (!_atspi_dbus_get_property (obj, atspi_interface_application, "ToolkitName", error, "s", &ret))
       return NULL;
-  return g_strdup (ret);
+  return ret;
 }
 
 /**
@@ -785,8 +782,9 @@ atspi_accessible_get_toolkit_version (AtspiAccessible *obj, GError **error)
 
   if (!_atspi_dbus_get_property (obj, atspi_interface_application, "ToolkitVersion", error, "s", &ret))
       return NULL;
-  return g_strdup (ret);
+  return ret;
 }
+
 /**
  * atspi_accessible_get_toolkit_version:
  * @obj: a pointer to the #AtspiAccessible object on which to operate.
@@ -1032,7 +1030,7 @@ atspi_accessible_is_streamable_content (AtspiAccessible *obj)
   return _atspi_accessible_is_a (obj,
 			      atspi_interface_streamable_content);
 #else
-  g_warning (_("Streamable content not implemented"));
+  g_warning ("Streamable content not implemented");
   return FALSE;
 #endif
 }
@@ -1303,9 +1301,6 @@ atspi_accessible_get_interfaces (AtspiAccessible *obj)
 {
   GArray *ret = g_array_new (TRUE, TRUE, sizeof (gchar *));
 
-  if (!ret)
-    return NULL;
-
   g_return_val_if_fail (obj != NULL, NULL);
 
   append_const_val (ret, "Accessible");
