@@ -28,6 +28,8 @@
 
 /* Private internal implementation details of at-spi. */
 
+#include <string.h>
+
 #include "atspi.h"
 
 #include "dbus/dbus.h"
@@ -181,6 +183,4 @@ typedef enum
   ATSPI_ERROR_APPLICATION_GONE,
   ATSPI_ERROR_IPC
 } AtspiError;
-
-extern GMainLoop *atspi_main_loop;
 #endif	/* _ATSPI_MISC_PRIVATE_H_ */
