@@ -4,6 +4,7 @@
  *
  * Copyright 2002 Ximian, Inc.
  *           2002 Sun Microsystems Inc.
+ * Copyright 2010, 2011 Novell, Inc.
  *           
  *
  * This library is free software; you can redistribute it and/or
@@ -33,4 +34,6 @@ void atspi_event_quit ();
 
 int atspi_exit ();
 
+DBusConnection *
+atspi_get_a11y_bus ();
 #endif	/* _ATSPI_MISC_H_ */
