@@ -24,12 +24,19 @@
 
 #include <config.h>
 #include <string.h>
-#include <dbus/dbus-glib-lowlevel.h>
+#include <ctype.h>
 
 #include "paths.h"
 #include "registry.h"
 #include "introspection.h"
 
+typedef struct event_data event_data;
+struct event_data
+{
+  gchar *bus_name;
+  gchar **data;
+};
+
 static void
 children_added_listener (DBusConnection * bus,
                          gint             index,
@@ -188,11 +195,83 @@ remove_application (SpiRegistry *reg, DBusConnection *bus, guint index)
 {
   SpiReference *ref = g_ptr_array_index (reg->apps, index);
 
-  /*TODO spi_remove_device_listeners (registry->de_controller, old);*/
+  spi_remove_device_listeners (reg->dec, ref->name);
   children_removed_listener (bus, index, ref->name, ref->path);
   g_ptr_array_remove_index (reg->apps, index);
 }
 
+static gboolean
+event_is_subtype (gchar **needle, gchar **haystack)
+{
+  while (*haystack && **haystack)
+    {
+      if (g_strcmp0 (*needle, *haystack))
+        return FALSE;
+      needle++;
+      haystack++;
+    }
+  return TRUE;
+}
+
+static gboolean
+needs_mouse_poll (char **event)
+{
+  if (g_strcmp0 (event [0], "Mouse") != 0)
+    return FALSE;
+  if (!event [1] || !event [1][0])
+    return TRUE;
+  return (g_strcmp0 (event [1], "Abs") == 0);
+}
+
+static void
+remove_events (SpiRegistry *registry, const char *bus_name, const char *event)
+{
+  event_data *evdata;
+  gchar **remove_data;
+  GList *list;
+  gboolean mouse_found = FALSE;
+  DBusMessage *signal;
+
+  remove_data = g_strsplit (event, ":", 3);
+  if (!remove_data)
+    {
+      return;
+    }
+
+  for (list = registry->events; list;)
+    {
+      event_data *evdata = list->data;
+      if (!g_strcmp0 (evdata->bus_name, bus_name) &&
+          event_is_subtype (evdata->data, remove_data))
+        {
+          list = list->next;
+          g_strfreev (evdata->data);
+          g_free (evdata->bus_name);
+          g_free (evdata);
+          registry->events = g_list_remove (registry->events, evdata);
+        }
+      else
+        {
+          if (needs_mouse_poll (evdata->data))
+            mouse_found = TRUE;
+          list = list->next;
+        }
+    }
+
+  if (!mouse_found)
+    spi_device_event_controller_stop_poll_mouse ();
+
+  g_strfreev (remove_data);
+
+  signal = dbus_message_new_signal (SPI_DBUS_PATH_REGISTRY,
+                                    SPI_DBUS_INTERFACE_REGISTRY,
+                                    "EventListenerDeregistered");
+  dbus_message_append_args (signal, DBUS_TYPE_STRING, &bus_name,
+                            DBUS_TYPE_STRING, &event, DBUS_TYPE_INVALID);
+  dbus_connection_send (registry->bus, signal, NULL);
+  dbus_message_unref (signal);
+}
+
 static void
 handle_disconnection (DBusConnection *bus, DBusMessage *message, void *user_data)
 {
@@ -218,19 +297,61 @@ handle_disconnection (DBusConnection *bus, DBusMessage *message, void *user_data
                   g_ptr_array_remove_index (reg->apps, i);
                 }
             } 
+
+          remove_events (reg, old, "");
         }
     }
 }
 
+/*
+ * Converts names of the form "active-descendant-changed" to
+ *" ActiveDescendantChanged"
+ */
+static gchar *
+ensure_proper_format (const char *name)
+{
+  gchar *ret = (gchar *) g_malloc (strlen (name) * 2 + 2);
+  gchar *p = ret;
+  gboolean need_upper = TRUE;
+
+  if (!ret)
+    return NULL;
+  while (*name)
+    {
+      if (need_upper)
+        {
+          *p++ = toupper (*name);
+          need_upper = FALSE;
+        }
+      else if (*name == '-')
+        need_upper = TRUE;
+      else if (*name == ':')
+        {
+          need_upper = TRUE;
+          *p++ = *name;
+        }
+      else
+        *p++ = *name;
+      name++;
+    }
+  *p = '\0';
+  return ret;
+}
+
 static DBusHandlerResult
 signal_filter (DBusConnection *bus, DBusMessage *message, void *user_data)
 {
   SpiRegistry *registry = SPI_REGISTRY (user_data);
   guint res = DBUS_HANDLER_RESULT_HANDLED;
+  const gint   type    = dbus_message_get_type (message);
   const char *iface = dbus_message_get_interface (message);
   const char *member = dbus_message_get_member (message);
 
-  if (!g_strcmp0(iface, DBUS_INTERFACE_DBUS) && !g_strcmp0(member, "NameOwnerChanged"))
+  if (type != DBUS_MESSAGE_TYPE_SIGNAL)
+    return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+
+  if (!g_strcmp0(iface, DBUS_INTERFACE_DBUS) &&
+      !g_strcmp0(member, "NameOwnerChanged"))
       handle_disconnection (bus, message, user_data);
   else
       res = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
@@ -469,8 +590,26 @@ impl_get_ChildCount (DBusMessageIter * iter, void *user_data)
 {
   SpiRegistry *reg = SPI_REGISTRY (user_data);
   dbus_int32_t rv = reg->apps->len;
+  dbus_bool_t result;
+  DBusMessageIter iter_variant;
 
-  return dbus_message_iter_append_basic (iter, DBUS_TYPE_INT32, &rv);
+  if (!dbus_message_iter_open_container (iter, DBUS_TYPE_VARIANT, "i",
+                                         &iter_variant))
+    return FALSE;
+  result = dbus_message_iter_append_basic (&iter_variant, DBUS_TYPE_INT32, &rv);
+  dbus_message_iter_close_container (iter, &iter_variant);
+  return result;
+}
+
+static dbus_bool_t
+impl_get_ToolkitName (DBusMessageIter * iter, void *user_data)
+{
+  return return_v_string (iter, "at-spi-registry");
+}
+
+impl_get_ToolkitVersion (DBusMessageIter * iter, void *user_data)
+{
+  return return_v_string (iter, "2.0");
 }
 
 static DBusMessage *
@@ -558,7 +697,7 @@ static DBusMessage *
 impl_GetRole (DBusConnection * bus, DBusMessage * message, void * user_data)
 {
   DBusMessage *reply;
-  dbus_uint32_t rv = 0;
+  dbus_uint32_t rv = 14;	/* TODO: Get DESKTOP_FRAME from somewhere */
 
   reply = dbus_message_new_method_return (message);
   dbus_message_append_args (reply, DBUS_TYPE_UINT32, &rv, DBUS_TYPE_INVALID);
@@ -570,7 +709,7 @@ impl_GetRoleName (DBusConnection * bus,
                   DBusMessage * message, void *user_data)
 {
   DBusMessage *reply;
-  const char *role_name = "unknown";
+  const char *role_name = "desktop frame";
 
   reply = dbus_message_new_method_return (message);
   dbus_message_append_args (reply, DBUS_TYPE_STRING, &role_name,
@@ -584,7 +723,7 @@ impl_GetLocalizedRoleName (DBusConnection * bus,
 {
   /* TODO - Localize this */
   DBusMessage *reply;
-  const char *role_name = "unknown";
+  const char *role_name = "desktop frame";
 
   reply = dbus_message_new_method_return (message);
   dbus_message_append_args (reply, DBUS_TYPE_STRING, &role_name,
@@ -667,6 +806,130 @@ impl_GetInterfaces (DBusConnection * bus,
   return reply;
 }
 
+static DBusMessage *
+impl_GetItems (DBusConnection * bus, DBusMessage * message, void *user_data)
+{
+  DBusMessage *reply;
+  DBusMessageIter iter, iter_array;
+
+  reply = dbus_message_new_method_return (message);
+
+  dbus_message_iter_init_append (reply, &iter);
+  dbus_message_iter_open_container (&iter, DBUS_TYPE_ARRAY,
+                                    "((so)(so)(so)a(so)assusau)", &iter_array);
+  dbus_message_iter_close_container (&iter, &iter_array);
+  return reply;
+}
+
+/* I would rather these two be signals, but I'm not sure that dbus-python
+ * supports emitting signals except for a service, so implementing as both
+ * a method call and signal for now.
+ */
+static DBusMessage *
+impl_register_event (DBusConnection *bus, DBusMessage *message, void *user_data)
+{
+  SpiRegistry *registry = SPI_REGISTRY (user_data);
+  const char *orig_name;
+  gchar *name;
+  event_data *evdata;
+  gchar **data;
+  GList *new_list;
+  DBusMessage *signal;
+  const char *sender = dbus_message_get_sender (message);
+
+  if (!dbus_message_get_args (message, NULL, DBUS_TYPE_STRING, &orig_name,
+    DBUS_TYPE_INVALID))
+    return NULL;
+
+  name = ensure_proper_format (orig_name);
+
+  evdata = (event_data *) g_malloc (sizeof (*evdata));
+  if (!evdata)
+    return NULL;
+  data = g_strsplit (name, ":", 3);
+  if (!data)
+    {
+      g_free (evdata);
+      return NULL;
+    }
+  if (!data [0])
+    data [1] = NULL;
+  if (!data [1])
+    data [2] = NULL;
+  evdata->bus_name = g_strdup (sender);
+  evdata->data = data;
+  new_list = g_list_append (registry->events, evdata);
+  if (new_list)
+    registry->events = new_list;
+
+  if (needs_mouse_poll (evdata->data))
+    {
+      spi_device_event_controller_start_poll_mouse (registry);
+    }
+
+  signal = dbus_message_new_signal (SPI_DBUS_PATH_REGISTRY,
+                                    SPI_DBUS_INTERFACE_REGISTRY,
+                                    "EventListenerRegistered");
+  dbus_message_append_args (signal, DBUS_TYPE_STRING, &sender,
+                            DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
+  dbus_connection_send (bus, signal, NULL);
+  dbus_message_unref (signal);
+
+  g_free (name);
+  return dbus_message_new_method_return (message);
+}
+
+static DBusMessage *
+impl_deregister_event (DBusConnection *bus, DBusMessage *message, void *user_data)
+{
+  SpiRegistry *registry = SPI_REGISTRY (user_data);
+  const char *orig_name;
+  gchar *name;
+  const char *sender = dbus_message_get_sender (message);
+
+  if (!dbus_message_get_args (message, NULL, DBUS_TYPE_STRING, &orig_name,
+    DBUS_TYPE_INVALID))
+    return NULL;
+  name = ensure_proper_format (orig_name);
+
+  remove_events (registry, sender, name);
+
+  g_free (name);
+  return dbus_message_new_method_return (message);
+}
+
+static DBusMessage *
+impl_get_registered_events (DBusConnection *bus, DBusMessage *message, void *user_data)
+{
+  SpiRegistry *registry = SPI_REGISTRY (user_data);
+  event_data *evdata;
+  DBusMessage *reply;
+  DBusMessageIter iter, iter_struct, iter_array;
+  GList *list;
+
+  reply = dbus_message_new_method_return (message);
+  if (!reply)
+    return NULL;
+
+  dbus_message_iter_init_append (reply, &iter);
+  dbus_message_iter_open_container (&iter, DBUS_TYPE_ARRAY, "(ss)", &iter_array);
+  for (list = registry->events; list; list = list->next)
+    {
+      gchar *str;
+      evdata = list->data;
+      str = g_strconcat (evdata->data [0],
+                         ":", (evdata->data [1]? evdata->data [1]: ""),
+                         ":", (evdata->data [2]? evdata->data [2]: ""), NULL);
+      dbus_message_iter_open_container (&iter_array, DBUS_TYPE_STRUCT, NULL, &iter_struct);
+      dbus_message_iter_append_basic (&iter_struct, DBUS_TYPE_STRING, &evdata->bus_name);
+      dbus_message_iter_append_basic (&iter_struct, DBUS_TYPE_STRING, &str);
+      dbus_message_iter_close_container (&iter_array, &iter_struct);
+      g_free (str);
+    }
+  dbus_message_iter_close_container (&iter, &iter_array);
+  return reply;
+}
+
 /*---------------------------------------------------------------------------*/
 
 static void 
@@ -696,7 +959,7 @@ static const char *introspection_footer =
 "</node>";
 
 static DBusMessage *
-impl_Introspect (DBusConnection * bus,
+impl_Introspect_root (DBusConnection * bus,
                  DBusMessage * message, void *user_data)
 {
   GString *output;
@@ -724,6 +987,34 @@ impl_Introspect (DBusConnection * bus,
   return reply;
 }
 
+static DBusMessage *
+impl_Introspect_registry (DBusConnection * bus,
+                 DBusMessage * message, void *user_data)
+{
+  GString *output;
+  gchar *final;
+  gint i;
+
+  const gchar *pathstr = SPI_DBUS_PATH_REGISTRY;
+
+  DBusMessage *reply;
+
+  output = g_string_new(introspection_header);
+
+  g_string_append_printf(output, introspection_node_element, pathstr);
+
+  g_string_append (output, spi_org_a11y_atspi_Registry);
+
+  g_string_append(output, introspection_footer);
+  final = g_string_free(output, FALSE);
+
+  reply = dbus_message_new_method_return (message);
+  dbus_message_append_args(reply, DBUS_TYPE_STRING, &final, DBUS_TYPE_INVALID);
+
+  g_free(final);
+  return reply;
+}
+
 /*---------------------------------------------------------------------------*/
 
 /*
@@ -806,7 +1097,7 @@ children_removed_listener (DBusConnection * bus,
 /*---------------------------------------------------------------------------*/
 
 static DBusHandlerResult
-handle_method (DBusConnection *bus, DBusMessage *message, void *user_data)
+handle_method_root (DBusConnection *bus, DBusMessage *message, void *user_data)
 {
   DBusHandlerResult result = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 
@@ -861,6 +1152,23 @@ handle_method (DBusConnection *bus, DBusMessage *message, void *user_data)
                       reply = dbus_message_new_error (message, DBUS_ERROR_FAILED, "Property unavailable");
                     }
                 }
+              else if (!strcmp (prop_iface, SPI_DBUS_INTERFACE_APPLICATION))
+                {
+                  if (!strcmp (prop_member, "ToolkitName"))
+                    impl_get_ToolkitName (&iter, user_data);
+                  else if (!strcmp (prop_member, "ToolkitVersion"))
+                    impl_get_ToolkitVersion (&iter, user_data);
+                  else
+                    {
+                      dbus_message_unref (reply); 
+                      reply = dbus_message_new_error (message, DBUS_ERROR_FAILED, "Property unavailable");
+                    }
+                }
+              else
+                {
+                  dbus_message_unref (reply); 
+                  reply = dbus_message_new_error (message, DBUS_ERROR_FAILED, "Property unavailable");
+                }
             }
           else
             {
@@ -943,7 +1251,105 @@ handle_method (DBusConnection *bus, DBusMessage *message, void *user_data)
     {
       result = DBUS_HANDLER_RESULT_HANDLED;
       if      (!strcmp (member, "Introspect"))
-          reply = impl_Introspect (bus, message, user_data);
+          reply = impl_Introspect_root (bus, message, user_data);
+      else
+          result = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+    }
+
+  if (result == DBUS_HANDLER_RESULT_HANDLED)
+    {
+      if (!reply)
+        {
+          reply = dbus_message_new_method_return (message);
+        }
+
+      dbus_connection_send (bus, reply, NULL);
+      dbus_message_unref (reply);
+    }
+#if 0
+  else
+    {
+      g_print ("Registry | Unhandled message : %s|%s\n", iface, member);
+    }
+#endif
+  
+  return result;
+}
+
+static DBusHandlerResult
+handle_method_cache (DBusConnection *bus, DBusMessage *message, void *user_data)
+{
+  DBusHandlerResult result = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+
+  const gchar *iface   = dbus_message_get_interface (message);
+  const gchar *member  = dbus_message_get_member (message);
+  const gint   type    = dbus_message_get_type (message);
+
+  DBusMessage *reply = NULL;
+
+  /* Check for basic reasons not to handle */
+  if (type   != DBUS_MESSAGE_TYPE_METHOD_CALL ||
+      member == NULL ||
+      iface  == NULL)
+      return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+
+  if (!strcmp (iface, SPI_DBUS_INTERFACE_CACHE))
+    {
+      result = DBUS_HANDLER_RESULT_HANDLED;
+      if      (!strcmp (member, "GetItems"))
+          reply = impl_GetItems (bus, message, user_data);
+      else
+         result = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+    }
+
+  if (result == DBUS_HANDLER_RESULT_HANDLED)
+    {
+      if (!reply)
+        {
+          reply = dbus_message_new_method_return (message);
+        }
+
+      dbus_connection_send (bus, reply, NULL);
+      dbus_message_unref (reply);
+    }
+  return result;
+}
+
+static DBusHandlerResult
+handle_method_registry (DBusConnection *bus, DBusMessage *message, void *user_data)
+{
+  DBusHandlerResult result = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+
+  const gchar *iface   = dbus_message_get_interface (message);
+  const gchar *member  = dbus_message_get_member (message);
+  const gint   type    = dbus_message_get_type (message);
+
+  DBusMessage *reply = NULL;
+
+  /* Check for basic reasons not to handle */
+  if (type   != DBUS_MESSAGE_TYPE_METHOD_CALL ||
+      member == NULL ||
+      iface  == NULL)
+      return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+
+  if (!strcmp (iface, SPI_DBUS_INTERFACE_REGISTRY))
+    {
+      result = DBUS_HANDLER_RESULT_HANDLED;
+      if (!strcmp(member, "RegisterEvent"))
+      reply = impl_register_event (bus, message, user_data);
+      else if (!strcmp(member, "DeregisterEvent"))
+        reply = impl_deregister_event (bus, message, user_data);
+      else if (!strcmp(member, "GetRegisteredEvents"))
+        reply = impl_get_registered_events (bus, message, user_data);
+      else
+          result = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+    }
+
+  if (!strcmp (iface, "org.freedesktop.DBus.Introspectable"))
+    {
+      result = DBUS_HANDLER_RESULT_HANDLED;
+      if      (!strcmp (member, "Introspect"))
+          reply = impl_Introspect_registry (bus, message, user_data);
       else
           result = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
     }
@@ -970,10 +1376,24 @@ handle_method (DBusConnection *bus, DBusMessage *message, void *user_data)
 
 /*---------------------------------------------------------------------------*/
 
+static DBusObjectPathVTable root_vtable =
+{
+  NULL,
+  &handle_method_root,
+  NULL, NULL, NULL, NULL
+};
+
 static DBusObjectPathVTable registry_vtable =
 {
   NULL,
-  &handle_method,
+  &handle_method_registry,
+  NULL, NULL, NULL, NULL
+};
+
+static DBusObjectPathVTable cache_vtable =
+{
+  NULL,
+  &handle_method_cache,
   NULL, NULL, NULL, NULL
 };
 
@@ -990,10 +1410,16 @@ spi_registry_new (DBusConnection *bus)
   dbus_bus_add_match (bus, app_sig_match_name_owner, NULL);
   dbus_connection_add_filter (bus, signal_filter, reg, NULL);
 
-  dbus_connection_register_object_path (bus, SPI_DBUS_PATH_ROOT, &registry_vtable, reg);
+  dbus_connection_register_object_path (bus, SPI_DBUS_PATH_ROOT, &root_vtable, reg);
+
+  dbus_connection_register_object_path (bus, SPI_DBUS_PATH_CACHE, &cache_vtable, reg);
+
+  dbus_connection_register_object_path (bus, SPI_DBUS_PATH_REGISTRY, &registry_vtable, reg);
 
   emit_Available (bus);
 
+  reg->events = NULL;
+
   return reg;
 }
 
